from decimal import Decimal
from typing import List

from degen.odds_models import Odds


class Bet(object):
    """Class representing a generic Bet"""

    def __init__(self, wager_amt: Decimal, odds: Odds):
        self.wager_amt = wager_amt
        self._odds = odds

    @property
    def american_odds(self):
        return self._odds.to_american_odds()

    @property
    def implied_probability(self):
        return self._odds.to_implied_probability()

    @property
    def decimal_odds(self):
        return self._odds.to_decimal_odds()

    @property
    def payoff(self):
        dec_odds = self.decimal_odds.value
        return self.wager_amt * dec_odds


class Parlay(Bet):
    def __init__(self, wager_amt: Decimal, legs: List[Odds]):
        self.legs = legs
        self.odds: Odds = legs[0]

        for odds in legs[1:]:
            self.odds &= odds

        super().__init__(wager_amt, self.odds)



from decimal import Decimal


def check_implied_probability(implied_probability: Decimal):
    if implied_probability > 1.0 or implied_probability < 0:
        raise ValueError("Implied probability must be between 0.0 and 1.0")



from decimal import Decimal

from degen.utils import check_implied_probability


def implied_prob_to_odds_pos(implied_prob: Decimal) -> Decimal:
    """Compute Positive American Odds given an implied prob"""
    check_implied_probability(implied_prob)
    return (Decimal(100.0) / implied_prob) - Decimal(100.0)


def implied_prob_to_odds_neg(implied_prob: Decimal) -> Decimal:
    """Compute Negative American Odds given an implied prob"""
    check_implied_probability(implied_prob)
    return (Decimal(100.0) * implied_prob) / (implied_prob - Decimal(1.0))


def american_odds_to_implied_prob(american_odds: Decimal) -> Decimal:
    """American odds to implied probability"""

    if american_odds == Decimal("-100"):
        american_odds = Decimal("100")

    if american_odds >= Decimal(100.0):
        return Decimal(100.0) / (american_odds + Decimal(100.0))
    else:
        return abs(american_odds) / (abs(american_odds) + Decimal(100.0))


def american_odds_to_decimal_odds(american_odds: Decimal) -> Decimal:
    """American odds to decimal odds"""
    if american_odds >= Decimal(100.0):
        return Decimal(1.0) + (american_odds / Decimal(100.0))
    else:
        return Decimal(1.0) - (Decimal(100.0) / american_odds)


def decimal_odds_to_american_odds(decimal_odds: Decimal) -> Decimal:
    """Decimal odds to american odds"""
    if decimal_odds >= Decimal(2.0):
        return (decimal_odds - Decimal(1.0)) * Decimal(100.0)
    else:
        return Decimal(-100.0) / (decimal_odds - Decimal(1.0))


def decimal_odds_to_implied_prob(decimal_odds: Decimal) -> Decimal:
    """Decimal odds to implied probability"""
    return Decimal(1.0) / decimal_odds


def implied_prob_to_decimal_odds(implied_prob: Decimal) -> Decimal:
    """Implied Probability to decimal odds"""
    return Decimal(1.0) / implied_prob



from pathlib import Path
from setuptools import setup

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

version = (this_directory / "__version__.txt").read_text()

setup(
    name="degen",
    version=version,
    description="Programming with betting odds, made simple",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Joey Zuhusky",
    keywords="odds sportsbetting betting python betting gambling",
    author_email="jzuhusky@gmail.com",
    maintainer_email="jzuhusky@gmail.com",
    url="https://github.com/jzuhusky/degen",
    install_requires=["pydantic>=1.9.1"],
    license="MIT",
    packages=["degen"],
)



import abc
from decimal import Decimal

from pydantic import BaseModel, validator

from degen.math_tools import (
    american_odds_to_implied_prob as aotip,
    american_odds_to_decimal_odds as aotdo,
    implied_prob_to_odds_neg as ipton,
    implied_prob_to_odds_pos as iptop,
    implied_prob_to_decimal_odds as iptdo,
    decimal_odds_to_american_odds as dotao,
    decimal_odds_to_implied_prob as dotip,
)


class Odds(BaseModel, abc.ABC):

    value: Decimal

    @abc.abstractmethod
    def to_american_odds(self):
        raise NotImplementedError

    @abc.abstractmethod
    def to_implied_probability(self):
        raise NotImplementedError

    @abc.abstractmethod
    def to_decimal_odds(self):
        raise NotImplementedError

    @property
    def is_juiced(self):
        return False

    def __eq__(self, other):
        return self.value == other.value

    def __or__(self, other):
        """Compute sum of implied odds and return new odds"""
        implied_odds_value = self.to_implied_probability().value
        other_implied_odds_value = other.to_implied_probability().value
        new_odds = ImpliedProbability(implied_odds_value + other_implied_odds_value)

        if new_odds.value >= Decimal(1.0):
            # only return implied prob - as these are juiced odds
            return new_odds

        if isinstance(self, AmericanOdds):
            return new_odds.to_american_odds()
        elif isinstance(self, DecimalOdds):
            return new_odds.to_decimal_odds()
        elif isinstance(self, ImpliedProbability):
            return new_odds
        raise TypeError()

    def __and__(self, other):
        """Compute product of implied odds and return new odds"""
        implied_odds_value = self.to_implied_probability().value
        other_implied_odds_value = other.to_implied_probability().value
        new_odds = ImpliedProbability(implied_odds_value * other_implied_odds_value)
        if isinstance(self, AmericanOdds):
            return new_odds.to_american_odds()
        elif isinstance(self, DecimalOdds):
            return new_odds.to_decimal_odds()
        elif isinstance(self, ImpliedProbability):
            return new_odds
        raise TypeError()


class AmericanOdds(Odds):

    # American odds are generally represented by integers >= 100 (+100)
    # or <= -100. But there is no reason why they *must* be integers, for this
    # reason, we allow them to be Decimals.
    # value: Decimal

    def __init__(self, value, **kwargs):
        super().__init__(value=value, **kwargs)

    @validator("value")
    def valid_american_odds(cls, v):
        if -100 < v < 100:
            raise ValueError("Invalid american odds. Odds must be >= +100 or <= -100")
        return v

    def to_american_odds(self):
        return self

    def to_decimal_odds(self):
        return DecimalOdds(value=aotdo(self.value))

    def to_implied_probability(self):
        return ImpliedProbability(value=aotip(self.value))


class DecimalOdds(Odds):
    def __init__(self, value, **kwargs):
        super().__init__(value=value, **kwargs)

    @validator("value")
    def valid_decimal_odds(cls, v):
        if v <= Decimal(0.0):
            raise ValueError("Decimal odds must be greater than 0.0")
        return v

    def to_american_odds(self):
        if self.is_juiced:
            raise ValueError("Cannot convert to juiced odds to american odds")
        return AmericanOdds(value=dotao(self.value))

    def to_decimal_odds(self):
        return self

    def to_implied_probability(self):
        return ImpliedProbability(value=dotip(self.value))

    @property
    def is_juiced(self):
        return self.value < Decimal(1.0)


class ImpliedProbability(Odds):
    def __init__(self, value, **kwargs):
        super().__init__(value=value, **kwargs)

    @validator("value")
    def valid_implied_odds(cls, v):
        if v < Decimal(0.0):
            # for betting purposes, it's sometimes useful to see implied
            # odds greater than 1.0. This is the "juice".
            raise ValueError("Implied odds must be greater than 0.0")
        return v

    def to_american_odds(self):
        if self.is_juiced:
            raise ValueError("Cannot convert to juiced odds to american odds")
        if self.value > Decimal(0.5):
            return AmericanOdds(value=ipton(self.value))
        return AmericanOdds(value=iptop(self.value))

    def to_decimal_odds(self):
        return DecimalOdds(value=iptdo(self.value))

    def to_implied_probability(self):
        return self

    @property
    def is_juiced(self):
        return self.value > Decimal(1.0)
